@@ -207,6 +207,9 @@ class Gem::TestCase < MiniTest::Unit::TestCase
     Gem.pre_uninstall do |uninstaller|
       @pre_uninstall_hook_arg = uninstaller
     end
+
+    @orig_yamler = YAML::ENGINE.yamler
+    YAML::ENGINE.yamler = 'psych' rescue nil
   end
 
   ##
@@ -238,12 +241,14 @@ class Gem::TestCase < MiniTest::Unit::TestCase
     else
       ENV.delete 'HOME'
     end
+
+    YAML::ENGINE.yamler = @orig_yamler
   end
 
   ##
   # Builds and installs the Gem::Specification +spec+
 
-  def install_gem spec
+  def install_gem spec, options = {}
     require 'rubygems/installer'
 
     use_ui Gem::MockGemUi.new do
@@ -254,7 +259,14 @@ class Gem::TestCase < MiniTest::Unit::TestCase
 
     gem = File.join(@tempdir, File.basename(spec.cache_file)).untaint
 
-    Gem::Installer.new(gem, :wrappers => true).install
+    Gem::Installer.new(gem, options.merge({:wrappers => true})).install
+  end
+
+  ##
+  # Builds and installs the Gem::Specification +spec+ into the user dir
+
+  def install_gem_user spec
+    install_gem spec, :user_install => true
   end
 
   ##
@@ -480,8 +492,11 @@ class Gem::TestCase < MiniTest::Unit::TestCase
 
     if deps then
       block = proc do |s|
-        deps.each do |n, req|
-          s.add_dependency n, (req || '>= 0')
+        # Since Hash#each is unordered in 1.8, sort
+        # the keys and iterate that way so the tests are
+        # deteriminstic on all implementations.
+        deps.keys.sort.each do |n|
+          s.add_dependency n, (deps[n] || '>= 0')
         end
       end
     end
@@ -501,8 +516,11 @@ class Gem::TestCase < MiniTest::Unit::TestCase
 
     if deps then
       block = proc do |s|
-        deps.each do |n, req|
-          s.add_dependency n, (req || '>= 0')
+        # Since Hash#each is unordered in 1.8, sort
+        # the keys and iterate that way so the tests are
+        # deteriminstic on all implementations.
+        deps.keys.sort.each do |n|
+          s.add_dependency n, (deps[n] || '>= 0')
         end
       end
     end
@@ -678,12 +696,13 @@ Also, a list:
     end
 
     v = Gem.marshal_version
+
     Gem::Specification.each do |spec|
       path = "#{@gem_repo}quick/Marshal.#{v}/#{spec.original_name}.gemspec.rz"
       data = Marshal.dump spec
       data_deflate = Zlib::Deflate.deflate data
       @fetcher.data[path] = data_deflate
-    end
+    end unless Gem::RemoteFetcher === @fetcher # HACK for test_download_to_cache
 
     nil # force errors
   end
@@ -854,4 +873,3 @@ Also, a list:
   end
 
 end
-
