@@ -43,7 +43,7 @@ rb_get_expanded_load_path(void)
 
     ary = rb_ary_new2(RARRAY_LEN(load_path));
     for (i = 0; i < RARRAY_LEN(load_path); ++i) {
-	VALUE path = rb_file_expand_path(RARRAY_PTR(load_path)[i], Qnil);
+	VALUE path = rb_file_expand_path_fast(RARRAY_PTR(load_path)[i], Qnil);
 	rb_str_freeze(path);
 	rb_ary_push(ary, path);
     }
@@ -77,14 +77,15 @@ loaded_feature_path(const char *name, long vlen, const char *feature, long len,
     long plen;
     const char *e;
 
-    if(vlen < len) return 0;
-    if (!strncmp(name+(vlen-len),feature,len)){
+    if (vlen < len) return 0;
+    if (!strncmp(name+(vlen-len), feature, len)) {
 	plen = vlen - len - 1;
-    } else {
+    }
+    else {
 	for (e = name + vlen; name != e && *e != '.' && *e != '/'; --e);
-	if (*e!='.' ||
+	if (*e != '.' ||
 	    e-name < len ||
-	    strncmp(e-len,feature,len) )
+	    strncmp(e-len, feature, len))
 	    return 0;
 	plen = e - name - len - 1;
     }
@@ -93,7 +94,7 @@ loaded_feature_path(const char *name, long vlen, const char *feature, long len,
 	const char *s = StringValuePtr(p);
 	long n = RSTRING_LEN(p);
 
-	if (n != plen ) continue;
+	if (n != plen) continue;
 	if (n && (strncmp(name, s, n) || name[n] != '/')) continue;
 	switch (type) {
 	  case 's':
@@ -183,7 +184,7 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
 	    fs.name = feature;
 	    fs.len = len;
 	    fs.type = type;
-	    fs.load_path = load_path ? load_path : rb_get_load_path();
+	    fs.load_path = load_path ? load_path : rb_get_expanded_load_path();
 	    fs.result = 0;
 	    st_foreach(loading_tbl, loaded_feature_path_i, (st_data_t)&fs);
 	    if ((f = fs.result) != 0) {
@@ -233,7 +234,7 @@ rb_feature_provided(const char *feature, const char **loading)
 
     if (*feature == '.' &&
 	(feature[1] == '/' || strncmp(feature+1, "./", 2) == 0)) {
-	fullpath = rb_file_expand_path(rb_str_new2(feature), Qnil);
+	fullpath = rb_file_expand_path_fast(rb_str_new2(feature), Qnil);
 	feature = RSTRING_PTR(fullpath);
     }
     if (ext && !strchr(ext, '/')) {
@@ -319,7 +320,8 @@ rb_load_internal(VALUE fname, int wrap)
     th->top_self = self;
     th->top_wrapper = wrapper;
 
-    if (!loaded) {
+    if (!loaded && !FIXNUM_P(GET_THREAD()->errinfo)) {
+	/* an error on loading don't include INT2FIX(TAG_FATAL) see r35625 */
 	rb_exc_raise(GET_THREAD()->errinfo);
     }
     if (state) {
@@ -397,7 +399,7 @@ load_lock(const char *ftptr)
 	}
 	/* partial state */
 	ftptr = ruby_strdup(ftptr);
-	data = (st_data_t)rb_barrier_new();
+	data = (st_data_t)rb_thread_shield_new();
 	st_insert(loading_tbl, (st_data_t)ftptr, data);
 	return (char *)ftptr;
     }
@@ -405,7 +407,28 @@ load_lock(const char *ftptr)
 	rb_warning("loading in progress, circular require considered harmful - %s", ftptr);
 	rb_backtrace();
     }
-    return RTEST(rb_barrier_wait((VALUE)data)) ? (char *)ftptr : 0;
+    switch (rb_thread_shield_wait((VALUE)data)) {
+      case Qfalse:
+	data = (st_data_t)ftptr;
+	st_delete(loading_tbl, &data, 0);
+	return 0;
+      case Qnil:
+	return 0;
+    }
+    return (char *)ftptr;
+}
+
+static int
+release_thread_shield(st_data_t *key, st_data_t *value, st_data_t done, int existing)
+{
+    VALUE thread_shield = (VALUE)*value;
+    if (!existing) return ST_STOP;
+    if (done ? rb_thread_shield_destroy(thread_shield) : rb_thread_shield_release(thread_shield)) {
+	/* still in-use */
+	return ST_CONTINUE;
+    }
+    xfree((char *)*key);
+    return ST_DELETE;
 }
 
 static void
@@ -413,41 +436,47 @@ load_unlock(const char *ftptr, int done)
 {
     if (ftptr) {
 	st_data_t key = (st_data_t)ftptr;
-	st_data_t data;
 	st_table *loading_tbl = get_loading_table();
 
-	if (st_delete(loading_tbl, &key, &data)) {
-	    VALUE barrier = (VALUE)data;
-	    xfree((char *)key);
-	    if (done)
-		rb_barrier_destroy(barrier);
-	    else
-		rb_barrier_release(barrier);
-	}
+	st_update(loading_tbl, key, release_thread_shield, done);
     }
 }
 
 
 /*
  *  call-seq:
- *     require(string)    -> true or false
+ *     require(name)    -> true or false
  *
- *  Ruby tries to load the library named _string_, returning
- *  +true+ if successful. If the filename does not resolve to
- *  an absolute path, it will be searched for in the directories listed
- *  in <code>$:</code>. If the file has the extension ``.rb'', it is
- *  loaded as a source file; if the extension is ``.so'', ``.o'', or
- *  ``.dll'', or whatever the default shared library extension is on
- *  the current platform, Ruby loads the shared library as a Ruby
- *  extension. Otherwise, Ruby tries adding ``.rb'', ``.so'', and so on
- *  to the name. The name of the loaded feature is added to the array in
- *  <code>$"</code>. A feature will not be loaded if its name already
- *  appears in <code>$"</code>. The file name is converted to an absolute
- *  path, so ``<code>require 'a'; require './a'</code>'' will not load
- *  <code>a.rb</code> twice.
+ *  Loads the given +name+, returning +true+ if successful and +false+ if the
+ *  feature is already loaded.
+ *
+ *  If the filename does not resolve to an absolute path, it will be searched
+ *  for in the directories listed in <code>$LOAD_PATH</code> (<code>$:</code>).
+ *
+ *  If the filename has the extension ".rb", it is loaded as a source file; if
+ *  the extension is ".so", ".o", or ".dll", or the default shared library
+ *  extension on the current platform, Ruby loads the shared library as a
+ *  Ruby extension.  Otherwise, Ruby tries adding ".rb", ".so", and so on
+ *  to the name until found.  If the file named cannot be found, a LoadError
+ *  will be raised.
+ *
+ *  For Ruby extensions the filename given may use any shared library
+ *  extension.  For example, on Linux the socket extension is "socket.so" and
+ *  <code>require 'socket.dll'</code> will load the socket extension.
+ *
+ *  The absolute path of the loaded file is added to
+ *  <code>$LOADED_FEATURES</code> (<code>$"</code>).  A file will not be
+ *  loaded again if its path already appears in <code>$"</code>.  For example,
+ *  <code>require 'a'; require './a'</code> will not load <code>a.rb</code>
+ *  again.
+ *
+ *    require "my-library.rb"
+ *    require "db-driver"
+ *
+ *  Any constants or globals within the loaded source file will be available
+ *  in the calling program's global namespace. However, local variables will
+ *  not be propagated to the loading environment.
  *
- *     require "my-library.rb"
- *     require "db-driver"
  */
 
 VALUE
@@ -469,7 +498,7 @@ rb_f_require_relative(VALUE obj, VALUE fname)
 {
     VALUE base = rb_current_realfilepath();
     if (NIL_P(base)) {
-	rb_raise(rb_eLoadError, "cannot infer basepath");
+	rb_loaderror("cannot infer basepath");
     }
     base = rb_file_dirname(base);
     return rb_require_safe(rb_file_absolute_path(fname, base), rb_safe_level());
@@ -488,7 +517,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
     if (ext && !strchr(ext, '/')) {
 	if (IS_RBEXT(ext)) {
 	    if (rb_feature_p(ftptr, ext, TRUE, FALSE, &loading)) {
-		if (loading) *path = rb_str_new2(loading);
+		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 'r';
 	    }
 	    if ((tmp = rb_find_file_safe(fname, safe_level)) != 0) {
@@ -501,10 +530,10 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 	}
 	else if (IS_SOEXT(ext)) {
 	    if (rb_feature_p(ftptr, ext, FALSE, FALSE, &loading)) {
-		if (loading) *path = rb_str_new2(loading);
+		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 's';
 	    }
-	    tmp = rb_str_new(RSTRING_PTR(fname), ext - RSTRING_PTR(fname));
+	    tmp = rb_str_subseq(fname, 0, ext - RSTRING_PTR(fname));
 #ifdef DLEXT2
 	    OBJ_FREEZE(tmp);
 	    if (rb_find_file_ext_safe(&tmp, loadable_ext + 1, safe_level)) {
@@ -526,7 +555,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 	}
 	else if (IS_DLEXT(ext)) {
 	    if (rb_feature_p(ftptr, ext, FALSE, FALSE, &loading)) {
-		if (loading) *path = rb_str_new2(loading);
+		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 's';
 	    }
 	    if ((tmp = rb_find_file_safe(fname, safe_level)) != 0) {
@@ -538,7 +567,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 	}
     }
     else if ((ft = rb_feature_p(ftptr, 0, FALSE, FALSE, &loading)) == 'r') {
-	if (loading) *path = rb_str_new2(loading);
+	if (loading) *path = rb_filesystem_str_new_cstr(loading);
 	return 'r';
     }
     tmp = fname;
@@ -565,9 +594,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 static void
 load_failed(VALUE fname)
 {
-    VALUE mesg = rb_str_buf_new_cstr("cannot load such file -- ");
-    rb_str_append(mesg, fname);	/* should be ASCII compatible */
-    rb_exc_raise(rb_exc_new3(rb_eLoadError, mesg));
+    rb_load_fail(fname, "cannot load such file");
 }
 
 static VALUE
@@ -657,11 +684,12 @@ init_ext_call(VALUE arg)
 RUBY_FUNC_EXPORTED void
 ruby_init_ext(const char *name, void (*init)(void))
 {
-    if (load_lock(name)) {
+    char* const lock_key = load_lock(name);
+    if (lock_key) {
 	rb_vm_call_cfunc(rb_vm_top_self(), init_ext_call, (VALUE)init,
 			 0, rb_str_new2(name));
 	rb_provide(name);
-	load_unlock(name, 1);
+	load_unlock(lock_key, 1);
     }
 }
 
@@ -705,7 +733,11 @@ rb_mod_autoload(VALUE mod, VALUE sym, VALUE file)
 static VALUE
 rb_mod_autoload_p(VALUE mod, VALUE sym)
 {
-    return rb_autoload_p(mod, rb_to_id(sym));
+    ID id = rb_check_id(&sym);
+    if (!id) {
+	return Qnil;
+    }
+    return rb_autoload_p(mod, id);
 }
 
 /*
