@@ -2,7 +2,7 @@
 
   ruby.c -
 
-  $Author: akr $
+  $Author: nobu $
   created at: Tue Aug 10 12:47:31 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -17,6 +17,7 @@
 #endif
 #include "ruby/ruby.h"
 #include "ruby/encoding.h"
+#include "internal.h"
 #include "eval_intern.h"
 #include "dln.h"
 #include <stdio.h>
@@ -51,11 +52,13 @@
 char *getenv();
 #endif
 
-VALUE rb_parser_get_yydebug(VALUE);
-VALUE rb_parser_set_yydebug(VALUE, VALUE);
+#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
 
-const char *ruby_get_inplace_mode(void);
-void ruby_set_inplace_mode(const char *);
+#if defined DISABLE_RUBYGEMS && DISABLE_RUBYGEMS
+#define DEFAULT_RUBYGEMS_ENABLED "disabled"
+#else
+#define DEFAULT_RUBYGEMS_ENABLED "enabled"
+#endif
 
 #define DISABLE_BIT(bit) (1U << disable_##bit)
 enum disable_flag_bits {
@@ -67,10 +70,14 @@ enum disable_flag_bits {
 #define DUMP_BIT(bit) (1U << dump_##bit)
 enum dump_flag_bits {
     dump_version,
+    dump_version_v,
     dump_copyright,
     dump_usage,
+    dump_help,
     dump_yydebug,
     dump_syntax,
+    dump_parsetree,
+    dump_parsetree_with_comment,
     dump_insns,
     dump_flag_count
 };
@@ -109,61 +116,90 @@ cmdline_options_init(struct cmdline_options *opt)
     opt->src.enc.index = src_encoding_index;
     opt->ext.enc.index = -1;
     opt->intern.enc.index = -1;
+#if defined DISABLE_RUBYGEMS && DISABLE_RUBYGEMS
+    opt->disable |= DISABLE_BIT(gems);
+#endif
     return opt;
 }
 
-static NODE *load_file(VALUE, const char *, int, struct cmdline_options *);
+static NODE *load_file(VALUE, VALUE, int, struct cmdline_options *);
 static void forbid_setid(const char *, struct cmdline_options *);
-#define forbid_setid(s) forbid_setid(s, opt)
+#define forbid_setid(s) forbid_setid((s), opt)
 
 static struct {
     int argc;
     char **argv;
-#if !defined(PSTAT_SETCMD) && !defined(HAVE_SETPROCTITLE)
-    size_t len;
-#endif
 } origarg;
 
 static void
-usage(const char *name)
+usage(const char *name, int help)
 {
     /* This message really ought to be max 23 lines.
      * Removed -h because the user already knows that option. Others? */
 
-    static const char *const usage_msg[] = {
-	"-0[octal]       specify record separator (\\0, if no argument)",
-	"-a              autosplit mode with -n or -p (splits $_ into $F)",
-	"-c              check syntax only",
-	"-Cdirectory     cd to directory, before executing your script",
-	"-d              set debugging flags (set $DEBUG to true)",
-	"-e 'command'    one line of script. Several -e's allowed. Omit [programfile]",
-	"-Eex[:in]       specify the default external and internal character encodings",
-	"-Fpattern       split() pattern for autosplit (-a)",
-	"-i[extension]   edit ARGV files in place (make backup if extension supplied)",
-	"-Idirectory     specify $LOAD_PATH directory (may be used more than once)",
-	"-l              enable line ending processing",
-	"-n              assume 'while gets(); ... end' loop around your script",
-	"-p              assume loop like -n but print line also like sed",
-	"-rlibrary       require the library, before executing your script",
-	"-s              enable some switch parsing for switches after script name",
-	"-S              look for the script using PATH environment variable",
-	"-T[level]       turn on tainting checks",
-	"-v              print version number, then turn on verbose mode",
-	"-w              turn warnings on for your script",
-	"-W[level]       set warning level; 0=silence, 1=medium, 2=verbose (default for level)",
-	"-x[directory]   strip off text before #!ruby line and perhaps cd to directory",
-	"--copyright     print the copyright",
-	"--version       print the version",
-	NULL
+    struct message {
+	const char *str;
+	unsigned short namelen, secondlen;
     };
-    const char *const *p = usage_msg;
+#define M(shortopt, longopt, desc) { \
+    shortopt " " longopt " " desc, \
+    (unsigned short)sizeof(shortopt), \
+    (unsigned short)sizeof(longopt), \
+}
+    static const struct message usage_msg[] = {
+	M("-0[octal]",	   "",			   "specify record separator (\\0, if no argument)"),
+	M("-a",		   "",			   "autosplit mode with -n or -p (splits $_ into $F)"),
+	M("-c",		   "",			   "check syntax only"),
+	M("-Cdirectory",   "",			   "cd to directory, before executing your script"),
+	M("-d",		   ", --debug",		   "set debugging flags (set $DEBUG to true)"),
+	M("-e 'command'",  "",			   "one line of script. Several -e's allowed. Omit [programfile]"),
+	M("-Eex[:in]",     ", --encoding=ex[:in]", "specify the default external and internal character encodings"),
+	M("-Fpattern",	   "",			   "split() pattern for autosplit (-a)"),
+	M("-i[extension]", "",			   "edit ARGV files in place (make backup if extension supplied)"),
+	M("-Idirectory",   "",			   "specify $LOAD_PATH directory (may be used more than once)"),
+	M("-l",		   "",			   "enable line ending processing"),
+	M("-n",		   "",			   "assume 'while gets(); ... end' loop around your script"),
+	M("-p",		   "",			   "assume loop like -n but print line also like sed"),
+	M("-rlibrary",	   "",			   "require the library, before executing your script"),
+	M("-s",		   "",			   "enable some switch parsing for switches after script name"),
+	M("-S",		   "",			   "look for the script using PATH environment variable"),
+	M("-T[level=1]",   "",			   "turn on tainting checks"),
+	M("-v",		   ", --verbose",	   "print version number, then turn on verbose mode"),
+	M("-w",		   "",			   "turn warnings on for your script"),
+	M("-W[level=2]",   "",			   "set warning level; 0=silence, 1=medium, 2=verbose"),
+	M("-x[directory]", "",			   "strip off text before #!ruby line and perhaps cd to directory"),
+	M("-h",		   "",			   "show this message, --help for more info"),
+    };
+    static const struct message help_msg[] = {
+	M("--copyright",                   "", "print the copyright"),
+	M("--enable=feature[,...]",        "", "enable features"),
+	M("--disable=feature[,...]",       "", "disable features"),
+	M("--internal-encoding=encoding",  "", "specify the default internal character encoding"),
+	M("--external-encoding=encoding",  "", "specify the default external character encoding"),
+	M("--version",                     "", "print the version"),
+	M("--help",			   "", "show this message, -h for short message"),
+    };
+    static const struct message features[] = {
+	M("gems",    "",        "rubygems (default: "DEFAULT_RUBYGEMS_ENABLED")"),
+	M("rubyopt", "",        "RUBYOPT environment variable (default: enabled)"),
+    };
+    int i, w = help ? 32 : 16, num = numberof(usage_msg) - (help ? 1 : 0);
+#define SHOW(m) printf("  %.*s%-*.*s%s\n", (m).namelen-1, (m).str, \
+		       w - (m).namelen + 1, (help ? (m).secondlen-1 : 0), (m).str + (m).namelen, \
+		       (m).str + (m).namelen + (m).secondlen)
 
     printf("Usage: %s [switches] [--] [programfile] [arguments]\n", name);
-    while (*p)
-	printf("  %s\n", *p++);
-}
+    for (i = 0; i < num; ++i)
+	SHOW(usage_msg[i]);
 
-VALUE rb_get_load_path(void);
+    if (!help) return;
+
+    for (i = 0; i < numberof(help_msg); ++i)
+	SHOW(help_msg[i]);
+    puts("Features:");
+    for (i = 0; i < numberof(features); ++i)
+	SHOW(features[i]);
+}
 
 #ifdef MANGLED_PATH
 static VALUE
@@ -253,7 +289,14 @@ push_include_cygwin(const char *path, VALUE (*filter)(VALUE))
 		p = strncpy(RSTRING_PTR(buf), p, len);
 	    }
 	}
-	if (cygwin_conv_to_posix_path(p, rubylib) == 0)
+#ifdef HAVE_CYGWIN_CONV_PATH
+#define CONV_TO_POSIX_PATH(p, lib) \
+	cygwin_conv_path(CCP_WIN_A_TO_POSIX|CCP_RELATIVE, (p), (lib), sizeof(lib))
+#else
+#define CONV_TO_POSIX_PATH(p, lib) \
+	cygwin_conv_to_posix_path((p), (lib))
+#endif
+	if (CONV_TO_POSIX_PATH(p, rubylib) == 0)
 	    p = rubylib;
 	push_include(p, filter);
 	if (!*s) break;
@@ -337,10 +380,11 @@ void
 ruby_init_loadpath_safe(int safe_level)
 {
     VALUE load_path;
+    ID id_initial_load_path_mark;
     extern const char ruby_initial_load_paths[];
     const char *paths = ruby_initial_load_paths;
 #if defined LOAD_RELATIVE
-# if defined HAVE_DLADDR || (defined __CYGWIN__ && defined CCP_WIN_A_TO_POSIX)
+# if defined HAVE_DLADDR || defined HAVE_CYGWIN_CONV_PATH
 #   define VARIABLE_LIBPATH 1
 # else
 #   define VARIABLE_LIBPATH 0
@@ -350,13 +394,13 @@ ruby_init_loadpath_safe(int safe_level)
     VALUE sopath;
 # else
     char libpath[MAXPATHLEN + 1];
-    size_t baselen;
 # endif
+    size_t baselen;
     char *p;
 
 #if defined _WIN32 || defined __CYGWIN__
 # if VARIABLE_LIBPATH
-    sopath = rb_str_tmp_new(MAXPATHLEN);
+    sopath = rb_str_new(0, MAXPATHLEN);
     libpath = RSTRING_PTR(sopath);
     GetModuleFileName(libruby, libpath, MAXPATHLEN);
 # else
@@ -366,12 +410,17 @@ ruby_init_loadpath_safe(int safe_level)
     _execname(libpath, sizeof(libpath) - 1);
 #elif defined(HAVE_DLADDR)
     Dl_info dli;
-    if (dladdr(expand_include_path, &dli)) {
-	VALUE fname = rb_str_new_cstr(dli.dli_fname);
-	sopath = rb_file_absolute_path(fname, Qnil);
-	rb_str_resize(fname, 0);
-	libpath = RSTRING_PTR(sopath);
+    if (dladdr((void *)(VALUE)expand_include_path, &dli)) {
+	char fbuf[MAXPATHLEN];
+	char *f = dln_find_file_r(dli.dli_fname, getenv(PATH_ENV), fbuf, sizeof(fbuf));
+	VALUE fname = rb_str_new_cstr(f ? f : dli.dli_fname);
+	rb_str_freeze(fname);
+	sopath = rb_realpath_internal(Qnil, fname, 1);
     }
+    else {
+	sopath = rb_str_new(0, 0);
+    }
+    libpath = RSTRING_PTR(sopath);
 #endif
 
 #if !VARIABLE_LIBPATH
@@ -385,7 +434,7 @@ ruby_init_loadpath_safe(int safe_level)
 	const int win_to_posix = CCP_WIN_A_TO_POSIX | CCP_RELATIVE;
 	size_t newsize = cygwin_conv_path(win_to_posix, libpath, 0, 0);
 	if (newsize > 0) {
-	    VALUE rubylib = rb_str_tmp_new(newsize);
+	    VALUE rubylib = rb_str_new(0, newsize);
 	    p = RSTRING_PTR(rubylib);
 	    if (cygwin_conv_path(win_to_posix, libpath, p, newsize) == 0) {
 		rb_str_resize(sopath, 0);
@@ -402,9 +451,21 @@ ruby_init_loadpath_safe(int safe_level)
 #endif
     p = strrchr(libpath, '/');
     if (p) {
+	static const char bindir[] = "/bin";
+#ifdef LIBDIR_BASENAME
+	static const char libdir[] = "/"LIBDIR_BASENAME;
+#else
+	static const char libdir[] = "/lib";
+#endif
+	const ptrdiff_t bindir_len = (ptrdiff_t)sizeof(bindir) - 1;
+	const ptrdiff_t libdir_len = (ptrdiff_t)sizeof(libdir) - 1;
 	*p = 0;
-	if (p - libpath > 3 && !(STRCASECMP(p - 4, "/bin") && strcmp(p - 4, "/lib"))) {
-	    p -= 4;
+	if (p - libpath >= bindir_len && !STRCASECMP(p - bindir_len, bindir)) {
+	    p -= bindir_len;
+	    *p = 0;
+	}
+	else if (p - libpath >= libdir_len && !strcmp(p - libdir_len, libdir)) {
+	    p -= libdir_len;
 	    *p = 0;
 	}
     }
@@ -413,32 +474,42 @@ ruby_init_loadpath_safe(int safe_level)
 	strlcpy(libpath, ".", sizeof(libpath));
 	p = libpath + 1;
     }
-
     baselen = p - libpath;
+#define PREFIX_PATH() rb_str_new(libpath, baselen)
+#else
+    baselen = p - libpath;
+    rb_str_resize(sopath, baselen);
+    libpath = RSTRING_PTR(sopath);
+#define PREFIX_PATH() sopath
+#endif
 
 #define BASEPATH() rb_str_buf_cat(rb_str_buf_new(baselen+len), libpath, baselen)
-#else
-    rb_str_set_len(sopath, p - libpath);
 
-#define BASEPATH() rb_str_dup(sopath)
-#endif
-
-#define RUBY_RELATIVE(path, len) rb_str_buf_cat(BASEPATH(), path, len)
+#define RUBY_RELATIVE(path, len) rb_str_buf_cat(BASEPATH(), (path), (len))
 #else
-#define RUBY_RELATIVE(path, len) rubylib_mangled_path(path, len)
+    static const char exec_prefix[] = RUBY_EXEC_PREFIX;
+#define RUBY_RELATIVE(path, len) rubylib_mangled_path((path), (len))
+#define PREFIX_PATH() RUBY_RELATIVE(exec_prefix, sizeof(exec_prefix)-1)
 #endif
-#define incpush(path) rb_ary_push(load_path, (path))
     load_path = GET_VM()->load_path;
 
     if (safe_level == 0) {
+#ifdef MANGLED_PATH
+	rubylib_mangled_path("", 0);
+#endif
 	ruby_push_include(getenv("RUBYLIB"), identical_path);
     }
 
+    id_initial_load_path_mark = rb_intern_const("@gem_prelude_index");
     while (*paths) {
 	size_t len = strlen(paths);
-	incpush(RUBY_RELATIVE(paths, len));
+	VALUE path = RUBY_RELATIVE(paths, len);
+	rb_ivar_set(path, id_initial_load_path_mark, path);
+	rb_ary_push(load_path, path);
 	paths += len + 1;
     }
+
+    rb_const_set(rb_cObject, rb_intern_const("TMP_RUBY_PREFIX"), rb_obj_freeze(PREFIX_PATH()));
 }
 
 
@@ -446,38 +517,52 @@ static void
 add_modules(VALUE *req_list, const char *mod)
 {
     VALUE list = *req_list;
+    VALUE feature;
 
     if (!list) {
 	*req_list = list = rb_ary_new();
 	RBASIC(list)->klass = 0;
     }
-    rb_ary_push(list, rb_obj_freeze(rb_str_new2(mod)));
+    feature = rb_str_new2(mod);
+    RBASIC(feature)->klass = 0;
+    rb_ary_push(list, feature);
 }
 
-extern void Init_ext(void);
-extern VALUE rb_vm_top_self(void);
-
 static void
 require_libraries(VALUE *req_list)
 {
     VALUE list = *req_list;
+    VALUE self = rb_vm_top_self();
     ID require;
     rb_thread_t *th = GET_THREAD();
-    rb_block_t *prev_base_block = th->base_block;
+    rb_encoding *extenc = rb_default_external_encoding();
     int prev_parse_in_eval = th->parse_in_eval;
-    th->base_block = 0;
     th->parse_in_eval = 0;
 
     Init_ext();		/* should be called here for some reason :-( */
     CONST_ID(require, "require");
     while (list && RARRAY_LEN(list) > 0) {
 	VALUE feature = rb_ary_shift(list);
-	rb_funcall2(rb_vm_top_self(), require, 1, &feature);
+	rb_enc_associate(feature, extenc);
+	RBASIC(feature)->klass = rb_cString;
+	OBJ_FREEZE(feature);
+	rb_funcall2(self, require, 1, &feature);
     }
     *req_list = 0;
 
     th->parse_in_eval = prev_parse_in_eval;
-    th->base_block = prev_base_block;
+}
+
+static rb_env_t*
+toplevel_context(void)
+{
+    rb_env_t *env;
+    VALUE toplevel_binding = rb_const_get(rb_cObject, rb_intern("TOPLEVEL_BINDING"));
+    rb_binding_t *bind;
+
+    GetBindingPtr(toplevel_binding, bind);
+    GetEnvPtr(bind->env, env);
+    return env;
 }
 
 static void
@@ -494,7 +579,7 @@ process_sflag(int *sflag)
 	    VALUE v = *args++;
 	    char *s = StringValuePtr(v);
 	    char *p;
-	    int hyphen = Qfalse;
+	    int hyphen = FALSE;
 
 	    if (s[0] != '-')
 		break;
@@ -511,7 +596,7 @@ process_sflag(int *sflag)
 		    break;
 		}
 		if (*p == '-') {
-		    hyphen = Qtrue;
+		    hyphen = TRUE;
 		}
 		else if (*p != '_' && !ISALNUM(*p)) {
 		    VALUE name_error[2];
@@ -544,8 +629,6 @@ process_sflag(int *sflag)
     }
 }
 
-NODE *rb_parser_append_print(VALUE, NODE *);
-NODE *rb_parser_while_loop(VALUE, NODE *, int, int);
 static long proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt);
 
 static void
@@ -596,16 +679,16 @@ moreswitches(const char *s, struct cmdline_options *opt, int envopt)
 }
 
 #define NAME_MATCH_P(name, str, len) \
-    ((len) < (int)sizeof(name) && strncmp((str), name, (len)) == 0)
+    ((len) < (int)sizeof(name) && strncmp((str), (name), (len)) == 0)
 
 #define UNSET_WHEN(name, bit, str, len)	\
-    if (NAME_MATCH_P(name, str, len)) { \
+    if (NAME_MATCH_P((name), (str), (len))) { \
 	*(unsigned int *)arg &= ~(bit); \
 	return;				\
     }
 
 #define SET_WHEN(name, bit, str, len)	\
-    if (NAME_MATCH_P(name, str, len)) { \
+    if (NAME_MATCH_P((name), (str), (len))) { \
 	*(unsigned int *)arg |= (bit);	\
 	return;				\
     }
@@ -643,11 +726,14 @@ dump_option(const char *str, int len, void *arg)
     SET_WHEN_DUMP(version);
     SET_WHEN_DUMP(copyright);
     SET_WHEN_DUMP(usage);
+    SET_WHEN_DUMP(help);
     SET_WHEN_DUMP(yydebug);
     SET_WHEN_DUMP(syntax);
+    SET_WHEN_DUMP(parsetree);
+    SET_WHEN_DUMP(parsetree_with_comment);
     SET_WHEN_DUMP(insns);
     rb_warn("don't know how to dump `%.*s',", len, str);
-    rb_warn("but only [version, copyright, usage, yydebug, syntax, insns].");
+    rb_warn("but only [version, copyright, usage, yydebug, syntax, parsetree, parsetree_with_comment, insns].");
 }
 
 static void
@@ -667,11 +753,11 @@ set_option_encoding_once(const char *type, VALUE *name, const char *e, long elen
 }
 
 #define set_internal_encoding_once(opt, e, elen) \
-    set_option_encoding_once("default_intenal", &opt->intern.enc.name, e, elen)
+    set_option_encoding_once("default_internal", &(opt)->intern.enc.name, (e), (elen))
 #define set_external_encoding_once(opt, e, elen) \
-    set_option_encoding_once("default_extenal", &opt->ext.enc.name, e, elen)
+    set_option_encoding_once("default_external", &(opt)->ext.enc.name, (e), (elen))
 #define set_source_encoding_once(opt, e, elen) \
-    set_option_encoding_once("source", &opt->src.enc.name, e, elen)
+    set_option_encoding_once("source", &(opt)->src.enc.name, (e), (elen))
 
 static long
 proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
@@ -692,17 +778,17 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 	switch (*s) {
 	  case 'a':
 	    if (envopt) goto noenvopt;
-	    opt->do_split = Qtrue;
+	    opt->do_split = TRUE;
 	    s++;
 	    goto reswitch;
 
 	  case 'p':
 	    if (envopt) goto noenvopt;
-	    opt->do_print = Qtrue;
+	    opt->do_print = TRUE;
 	    /* through */
 	  case 'n':
 	    if (envopt) goto noenvopt;
-	    opt->do_loop = Qtrue;
+	    opt->do_loop = TRUE;
 	    s++;
 	    goto reswitch;
 
@@ -723,7 +809,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 		s++;
 		goto reswitch;
 	    }
-	    ruby_show_version();
+	    opt->dump |= DUMP_BIT(version_v);
 	    opt->verbose = 1;
 	  case 'w':
 	    ruby_verbose = Qtrue;
@@ -775,7 +861,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 
 	  case 'l':
 	    if (envopt) goto noenvopt;
-	    opt->do_line = Qtrue;
+	    opt->do_line = TRUE;
 	    rb_output_rs = rb_rs;
 	    s++;
 	    goto reswitch;
@@ -783,7 +869,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 	  case 'S':
 	    if (envopt) goto noenvopt;
 	    forbid_setid("-S");
-	    opt->do_search = Qtrue;
+	    opt->do_search = TRUE;
 	    s++;
 	    goto reswitch;
 
@@ -825,7 +911,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 
 	  case 'x':
 	    if (envopt) goto noenvopt;
-	    opt->xflag = Qtrue;
+	    opt->xflag = TRUE;
 	    s++;
 	    if (*s && chdir(s) < 0) {
 		rb_fatal("Can't chdir to %s", s);
@@ -946,17 +1032,17 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 	    s++;
 
 #	define is_option_end(c, allow_hyphen) \
-	    (!(c) || (allow_hyphen && (c) == '-') || (c) == '=')
+	    (!(c) || ((allow_hyphen) && (c) == '-') || (c) == '=')
 #	define check_envopt(name, allow_envopt) \
-	    ((allow_envopt || !envopt) ? (void)0 : \
+	    (((allow_envopt) || !envopt) ? (void)0 : \
 	     rb_raise(rb_eRuntimeError, "invalid switch in RUBYOPT: --" name))
 #	define need_argument(name, s) \
-	    ((*s++ ? !*s : (!--argc || !(s = *++argv))) ?		\
+	    ((*(s)++ ? !*(s) : (!--argc || !((s) = *++argv))) ?		\
 	     rb_raise(rb_eRuntimeError, "missing argument for --" name) \
 	     : (void)0)
 #	define is_option_with_arg(name, allow_hyphen, allow_envopt) \
-	    (strncmp(name, s, n = sizeof(name) - 1) == 0 && is_option_end(s[n], allow_hyphen) ? \
-	     (check_envopt(name, allow_envopt), s += n, need_argument(name, s), 1) : 0)
+	    (strncmp((name), s, n = sizeof(name) - 1) == 0 && is_option_end(s[n], (allow_hyphen)) ? \
+	     (check_envopt(name, (allow_envopt)), s += n, need_argument(name, s), 1) : 0)
 
 	    if (strcmp("copyright", s) == 0) {
 		if (envopt) goto noenvopt_long;
@@ -988,7 +1074,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 		    if (!*(s = ++p)) break;
 		    set_encoding_part(internal);
 		    if (!*(s = ++p)) break;
-#if ALLOW_DEFAULT_SOURCE_ENCODING
+#if defined ALLOW_DEFAULT_SOURCE_ENCODING && ALLOW_DEFAULT_SOURCE_ENCODING
 		    set_encoding_part(source);
 		    if (!*(s = ++p)) break;
 #endif
@@ -1003,7 +1089,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 	    else if (is_option_with_arg("external-encoding", Qfalse, Qtrue)) {
 		set_external_encoding_once(opt, s, 0);
 	    }
-#if ALLOW_DEFAULT_SOURCE_ENCODING
+#if defined ALLOW_DEFAULT_SOURCE_ENCODING && ALLOW_DEFAULT_SOURCE_ENCODING
 	    else if (is_option_with_arg("source-encoding", Qfalse, Qtrue)) {
 		set_source_encoding_once(opt, s, 0);
 	    }
@@ -1025,7 +1111,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 	    }
 	    else if (strcmp("help", s) == 0) {
 		if (envopt) goto noenvopt_long;
-		opt->dump |= DUMP_BIT(usage);
+		opt->dump |= DUMP_BIT(help);
 		goto switch_end;
 	    }
 	    else {
@@ -1075,13 +1161,11 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
     return argc0 - argc;
 }
 
-void Init_prelude(void);
-
 static void
-ruby_init_gems(int enable)
+ruby_init_prelude(void)
 {
-    if (enable) rb_define_module("Gem");
     Init_prelude();
+    rb_const_remove(rb_cObject, rb_intern_const("TMP_RUBY_PREFIX"));
 }
 
 static int
@@ -1123,7 +1207,7 @@ uscore_get(void)
     VALUE line;
 
     line = rb_lastline_get();
-    if (TYPE(line) != T_STRING) {
+    if (!RB_TYPE_P(line, T_STRING)) {
 	rb_raise(rb_eTypeError, "$_ value need to be String (%s given)",
 		 NIL_P(line) ? "nil" : rb_obj_classname(line));
     }
@@ -1132,8 +1216,8 @@ uscore_get(void)
 
 /*
  *  call-seq:
- *     sub(pattern, replacement)   => $_
- *     sub(pattern) { block }      => $_
+ *     sub(pattern, replacement)   -> $_
+ *     sub(pattern) { block }      -> $_
  *
  *  Equivalent to <code>$_.sub(<i>args</i>)</code>, except that
  *  <code>$_</code> will be updated if substitution occurs.
@@ -1141,19 +1225,17 @@ uscore_get(void)
  */
 
 static VALUE
-rb_f_sub(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_sub(int argc, VALUE *argv)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("sub"), argc, argv);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("sub"), argc, argv);
     rb_lastline_set(str);
     return str;
 }
 
 /*
  *  call-seq:
- *     gsub(pattern, replacement)    => string
- *     gsub(pattern) {|...| block }  => string
+ *     gsub(pattern, replacement)    -> string
+ *     gsub(pattern) {|...| block }  -> string
  *
  *  Equivalent to <code>$_.gsub...</code>, except that <code>$_</code>
  *  receives the modified result.
@@ -1162,18 +1244,16 @@ rb_f_sub(argc, argv)
  */
 
 static VALUE
-rb_f_gsub(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_gsub(int argc, VALUE *argv)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("gsub"), argc, argv);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("gsub"), argc, argv);
     rb_lastline_set(str);
     return str;
 }
 
 /*
  *  call-seq:
- *     chop   => string
+ *     chop   -> string
  *
  *  Equivalent to <code>($_.dup).chop!</code>, except <code>nil</code>
  *  is never returned. See <code>String#chop!</code>.
@@ -1184,7 +1264,7 @@ rb_f_gsub(argc, argv)
 static VALUE
 rb_f_chop(void)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("chop"), 0, 0);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("chop"), 0, 0);
     rb_lastline_set(str);
     return str;
 }
@@ -1192,8 +1272,8 @@ rb_f_chop(void)
 
 /*
  *  call-seq:
- *     chomp            => $_
- *     chomp(string)    => $_
+ *     chomp            -> $_
+ *     chomp(string)    -> $_
  *
  *  Equivalent to <code>$_ = $_.chomp(<em>string</em>)</code>. See
  *  <code>String#chomp</code>.
@@ -1202,15 +1282,16 @@ rb_f_chop(void)
  */
 
 static VALUE
-rb_f_chomp(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_chomp(int argc, VALUE *argv)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("chomp"), argc, argv);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("chomp"), argc, argv);
     rb_lastline_set(str);
     return str;
 }
 
+/* blank function in dmyext.c or generated by enc/make_encmake.rb */
+extern void Init_enc(void);
+
 static VALUE
 process_options(int argc, char **argv, struct cmdline_options *opt)
 {
@@ -1227,8 +1308,8 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     argc -= i;
     argv += i;
 
-    if (opt->dump & DUMP_BIT(usage)) {
-	usage(origarg.argv[0]);
+    if (opt->dump & (DUMP_BIT(usage)|DUMP_BIT(help))) {
+	usage(origarg.argv[0], (opt->dump & DUMP_BIT(help)));
 	return Qtrue;
     }
 
@@ -1248,9 +1329,12 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	    opt->intern.enc.name = int_enc_name;
     }
 
-    if (opt->dump & DUMP_BIT(version)) {
+    if (opt->src.enc.name)
+	rb_warning("-K is specified; it is for 1.8 compatibility and may cause odd behavior");
+
+    if (opt->dump & (DUMP_BIT(version) | DUMP_BIT(version_v))) {
 	ruby_show_version();
-	return Qtrue;
+	if (opt->dump & DUMP_BIT(version)) return Qtrue;
     }
     if (opt->dump & DUMP_BIT(copyright)) {
 	ruby_show_copyright();
@@ -1295,12 +1379,13 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 #if defined DOSISH || defined __CYGWIN__
     translit_char(RSTRING_PTR(opt->script_name), '\\', '/');
 #endif
-    rb_obj_freeze(opt->script_name);
 
     ruby_init_loadpath_safe(opt->safe_level);
+    Init_enc();
     rb_enc_find_index("encdb");
     lenc = rb_locale_encoding();
     rb_enc_associate(rb_progname, lenc);
+    rb_obj_freeze(rb_progname);
     parser = rb_parser_new();
     if (opt->dump & DUMP_BIT(yydebug)) {
 	rb_parser_set_yydebug(parser, Qtrue);
@@ -1328,6 +1413,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	opt->intern.enc.index = -1;
     }
     rb_enc_associate(opt->script_name, lenc);
+    rb_obj_freeze(opt->script_name);
     {
 	long i;
 	VALUE load_path = GET_VM()->load_path;
@@ -1335,20 +1421,18 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	    rb_enc_associate(RARRAY_PTR(load_path)[i], lenc);
 	}
     }
-    ruby_init_gems(!(opt->disable & DISABLE_BIT(gems)));
-    rb_progname = opt->script_name;
-    rb_vm_set_progname(rb_progname);
+    if (!(opt->disable & DISABLE_BIT(gems))) {
+#if defined DISABLE_RUBYGEMS && DISABLE_RUBYGEMS
+	rb_require("rubygems");
+#else
+	rb_define_module("Gem");
+#endif
+    }
+    ruby_init_prelude();
     ruby_set_argv(argc, argv);
     process_sflag(&opt->sflag);
 
-    {
-	/* set eval context */
-	VALUE toplevel_binding = rb_const_get(rb_cObject, rb_intern("TOPLEVEL_BINDING"));
-	rb_binding_t *bind;
-
-	GetBindingPtr(toplevel_binding, bind);
-	GetEnvPtr(bind->env, env);
-    }
+    env = toplevel_context();
 
 #define PREPARE_PARSE_MAIN(expr) do { \
     th->parse_in_eval--; \
@@ -1359,6 +1443,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 } while (0)
 
     if (opt->e_script) {
+	VALUE progname = rb_progname;
 	rb_encoding *eenc;
 	if (opt->src.enc.index >= 0) {
 	    eenc = rb_enc_from_index(opt->src.enc.index);
@@ -1367,7 +1452,9 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	    eenc = lenc;
 	}
 	rb_enc_associate(opt->e_script, eenc);
+        ruby_set_script_name(opt->script_name);
 	require_libraries(&opt->req_list);
+        ruby_set_script_name(progname);
 
 	PREPARE_PARSE_MAIN({
 	    tree = rb_parser_compile_string(parser, opt->script, opt->e_script, 1);
@@ -1379,9 +1466,10 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	}
 
 	PREPARE_PARSE_MAIN({
-	    tree = load_file(parser, opt->script, 1, opt);
+	    tree = load_file(parser, opt->script_name, 1, opt);
 	});
     }
+    ruby_set_script_name(opt->script_name);
     if (opt->dump & DUMP_BIT(yydebug)) return Qtrue;
 
     if (opt->ext.enc.index >= 0) {
@@ -1399,6 +1487,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     else if (!rb_default_internal_encoding())
 	/* Freeze default_internal */
 	rb_enc_set_default_internal(Qnil);
+    rb_stdio_set_default_encoding();
 
     if (!tree) return Qfalse;
 
@@ -1430,8 +1519,17 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	rb_define_global_function("chomp", rb_f_chomp, -1);
     }
 
+    if (opt->dump & DUMP_BIT(parsetree) || opt->dump & DUMP_BIT(parsetree_with_comment)) {
+	rb_io_write(rb_stdout, rb_parser_dump_tree(tree, opt->dump & DUMP_BIT(parsetree_with_comment)));
+	rb_io_flush(rb_stdout);
+	return Qtrue;
+    }
+
     PREPARE_PARSE_MAIN({
-	iseq = rb_iseq_new_main(tree, opt->script_name);
+	VALUE path = Qnil;
+	if (!opt->e_script && strcmp(opt->script, "-"))
+	    path = rb_realpath_internal(Qnil, opt->script_name, 1);
+	iseq = rb_iseq_new_main(tree, opt->script_name, path);
     });
 
     if (opt->dump & DUMP_BIT(insns)) {
@@ -1445,13 +1543,14 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     rb_define_readonly_boolean("$-a", opt->do_split);
 
     rb_set_safe_level(opt->safe_level);
+    rb_gc_set_params();
 
     return iseq;
 }
 
 struct load_file_arg {
     VALUE parser;
-    const char *fname;
+    VALUE fname;
     int script;
     struct cmdline_options *opt;
 };
@@ -1462,7 +1561,8 @@ load_file_internal(VALUE arg)
     extern VALUE rb_stdin;
     struct load_file_arg *argp = (struct load_file_arg *)arg;
     VALUE parser = argp->parser;
-    const char *fname = argp->fname;
+    VALUE fname_v = rb_str_encode_ospath(argp->fname);
+    const char *fname = StringValueCStr(fname_v);
     int script = argp->script;
     struct cmdline_options *opt = argp->opt;
     VALUE f;
@@ -1470,9 +1570,8 @@ load_file_internal(VALUE arg)
     NODE *tree = 0;
     rb_encoding *enc;
     ID set_encoding;
+    int xflag = 0;
 
-    if (!fname)
-	rb_load_fail(fname);
     if (strcmp(fname, "-") == 0) {
 	f = rb_stdin;
     }
@@ -1481,14 +1580,27 @@ load_file_internal(VALUE arg)
 #if defined DOSISH || defined __CYGWIN__
 	{
 	    const char *ext = strrchr(fname, '.');
-	    if (ext && STRCASECMP(ext, ".exe") == 0)
+	    if (ext && STRCASECMP(ext, ".exe") == 0) {
 		mode |= O_BINARY;
+		xflag = 1;
+	    }
 	}
 #endif
-	if ((fd = open(fname, mode)) < 0) {
-	    rb_load_fail(fname);
+	if ((fd = rb_cloexec_open(fname, mode, 0)) < 0) {
+	    rb_load_fail(fname_v, strerror(errno));
 	}
-
+        rb_update_max_fd(fd);
+#if !defined DOSISH && !defined __CYGWIN__
+	{
+	    struct stat st;
+	    if (fstat(fd, &st) != 0)
+		rb_load_fail(fname_v, strerror(errno));
+	    if (S_ISDIR(st.st_mode)) {
+		errno = EISDIR;
+		rb_load_fail(fname_v, strerror(EISDIR));
+	    }
+	}
+#endif
 	f = rb_io_fdopen(fd, mode, fname);
     }
 
@@ -1504,9 +1616,11 @@ load_file_internal(VALUE arg)
 	enc = rb_ascii8bit_encoding();
 	rb_funcall(f, set_encoding, 1, rb_enc_from_encoding(enc));
 
-	if (opt->xflag) {
+	if (xflag || opt->xflag) {
+	    line_start--;
+	  search_shebang:
 	    forbid_setid("-x");
-	    opt->xflag = Qfalse;
+	    opt->xflag = FALSE;
 	    while (!NIL_P(line = rb_io_gets(f))) {
 		line_start++;
 		if (RSTRING_LEN(line) > 2
@@ -1517,7 +1631,7 @@ load_file_internal(VALUE arg)
 		    }
 		}
 	    }
-	    rb_raise(rb_eLoadError, "no Ruby script found in input");
+	    rb_loaderror("no Ruby script found in input");
 	}
 
 	c = rb_io_getbyte(f);
@@ -1529,37 +1643,8 @@ load_file_internal(VALUE arg)
 		    return 0;
 
 		if ((p = strstr(RSTRING_PTR(line), "ruby")) == 0) {
-		    void rb_thread_stop_timer_thread(void);
-		    /* not ruby script, kick the program */
-		    char **argv;
-		    char *path;
-		    char *pend = RSTRING_PTR(line) + RSTRING_LEN(line);
-
-		    p = RSTRING_PTR(line);	/* skip `#!' */
-		    if (pend[-1] == '\n')
-			pend--;	/* chomp line */
-		    if (pend[-1] == '\r')
-			pend--;
-		    *pend = '\0';
-		    while (p < pend && ISSPACE(*p))
-			p++;
-		    path = p;	/* interpreter path */
-		    while (p < pend && !ISSPACE(*p))
-			p++;
-		    *p++ = '\0';
-		    if (p < pend) {
-			argv = ALLOCA_N(char *, origarg.argc + 3);
-			argv[1] = p;
-			MEMCPY(argv + 2, origarg.argv + 1, char *, origarg.argc);
-		    }
-		    else {
-			argv = origarg.argv;
-		    }
-		    argv[0] = path;
-		    rb_thread_stop_timer_thread();
-		    execv(path, argv);
-
-		    rb_fatal("Can't exec %s", path);
+		    /* not ruby script, assume -x flag */
+		    goto search_shebang;
 		}
 
 	      start_read:
@@ -1592,6 +1677,11 @@ load_file_internal(VALUE arg)
 	else if (!NIL_P(c)) {
 	    rb_io_ungetbyte(f, c);
 	}
+	else {
+	    if (f != rb_stdin) rb_io_close(f);
+	    f = Qnil;
+	}
+        ruby_set_script_name(opt->script_name);
 	require_libraries(&opt->req_list);	/* Why here? unnatural */
     }
     if (opt->src.enc.index >= 0) {
@@ -1603,18 +1693,32 @@ load_file_internal(VALUE arg)
     else {
 	enc = rb_usascii_encoding();
     }
+    if (NIL_P(f)) {
+	f = rb_str_new(0, 0);
+	rb_enc_associate(f, enc);
+	return (VALUE)rb_parser_compile_string(parser, fname, f, line_start);
+    }
     rb_funcall(f, set_encoding, 2, rb_enc_from_encoding(enc), rb_str_new_cstr("-"));
     tree = rb_parser_compile_file(parser, fname, f, line_start);
     rb_funcall(f, set_encoding, 1, rb_parser_encoding(parser));
-    if (script && rb_parser_end_seen_p(parser)) {
+    if (script && tree && rb_parser_end_seen_p(parser)) {
+	/*
+	 * DATA is a File that contains the data section of the executed file.
+	 * To create a data section use <tt>__END__</tt>:
+	 *
+	 *   $ cat t.rb
+	 *   puts DATA.gets
+	 *   __END__
+	 *   hello world!
+	 *
+	 *   $ ruby t.rb
+	 *   hello world!
+	 */
 	rb_define_global_const("DATA", f);
     }
     else if (f != rb_stdin) {
 	rb_io_close(f);
     }
-    else {
-	rb_io_ungetbyte(f, Qnil);
-    }
     return (VALUE)tree;
 }
 
@@ -1625,7 +1729,7 @@ restore_lineno(VALUE lineno)
 }
 
 static NODE *
-load_file(VALUE parser, const char *fname, int script, struct cmdline_options *opt)
+load_file(VALUE parser, VALUE fname, int script, struct cmdline_options *opt)
 {
     struct load_file_arg arg;
     arg.parser = parser;
@@ -1639,53 +1743,11 @@ void *
 rb_load_file(const char *fname)
 {
     struct cmdline_options opt;
+    VALUE fname_v = rb_str_new_cstr(fname);
 
-    return load_file(rb_parser_new(), fname, 0, cmdline_options_init(&opt));
+    return load_file(rb_parser_new(), fname_v, 0, cmdline_options_init(&opt));
 }
 
-#if !defined(PSTAT_SETCMD) && !defined(HAVE_SETPROCTITLE)
-#if !defined(_WIN32) && !(defined(HAVE_SETENV) && defined(HAVE_UNSETENV))
-#define USE_ENVSPACE_FOR_ARG0
-#endif
-
-#ifdef USE_ENVSPACE_FOR_ARG0
-extern char **environ;
-#endif
-
-static size_t
-get_arglen(int argc, char **argv)
-{
-    char *s = argv[0];
-    int i;
-
-    if (!argc) return 0;
-    s += strlen(s);
-    /* See if all the arguments are contiguous in memory */
-    for (i = 1; i < argc; i++) {
-	if (argv[i] == s + 1) {
-	    s++;
-	    s += strlen(s);	/* this one is ok too */
-	}
-	else {
-	    break;
-	}
-    }
-#if defined(USE_ENVSPACE_FOR_ARG0)
-    if (environ && (s == environ[0])) {
-	s += strlen(s);
-	for (i = 1; environ[i]; i++) {
-	    if (environ[i] == s + 1) {
-		s++;
-		s += strlen(s);	/* this one is ok too */
-	    }
-	}
-	ruby_setenv("", NULL); /* duplicate environ vars */
-    }
-#endif
-    return s - argv[0];
-}
-#endif
-
 static void
 set_arg0(VALUE val, ID id)
 {
@@ -1697,54 +1759,37 @@ set_arg0(VALUE val, ID id)
     StringValue(val);
     s = RSTRING_PTR(val);
     i = RSTRING_LEN(val);
-#if defined(PSTAT_SETCMD)
-    if (i > PST_CLEN) {
-	union pstun un;
-	char buf[PST_CLEN + 1];	/* PST_CLEN is 64 (HP-UX 11.23) */
-	strlcpy(buf, s, sizeof(buf));
-	un.pst_command = buf;
-	pstat(PSTAT_SETCMD, un, PST_CLEN, 0, 0);
-    }
-    else {
-	union pstun un;
-	un.pst_command = s;
-	pstat(PSTAT_SETCMD, un, i, 0, 0);
-    }
-#elif defined(HAVE_SETPROCTITLE)
+
     setproctitle("%.*s", (int)i, s);
-#else
 
-    if ((size_t)i >= origarg.len) {
-	i = (long)(origarg.len - 1);
-    }
-
-    memcpy(origarg.argv[0], s, i);
-
-    {
-	int j;
-	char *t = origarg.argv[0] + i;
-	*t = '\0';
-
-	if ((size_t)(i + 1) < origarg.len) {
-	    memset(t + 1, ' ', origarg.len - i - 1);
-	}
-	for (j = 1; j < origarg.argc; j++) {
-	    origarg.argv[j] = t;
-	}
-    }
-#endif
     rb_progname = rb_obj_freeze(rb_external_str_new(s, i));
 }
 
+/*! Sets the current script name to this value.
+ *
+ * This is similiar to <code>$0 = name</code> in Ruby level but also affects
+ * <code>Method#location</code> and others.
+ */
 void
 ruby_script(const char *name)
 {
     if (name) {
-	rb_progname = rb_obj_freeze(rb_external_str_new(name, strlen(name)));
+	rb_progname = rb_external_str_new(name, strlen(name));
 	rb_vm_set_progname(rb_progname);
     }
 }
 
+/*! Sets the current script name to this value.
+ *
+ * Same as ruby_script() but accepts a VALUE.
+ */
+void
+ruby_set_script_name(VALUE name)
+{
+    rb_progname = rb_str_dup(name);
+    rb_vm_set_progname(rb_progname);
+}
+
 static void
 init_ids(struct cmdline_options *opt)
 {
@@ -1790,10 +1835,12 @@ opt_W_getter(ID id, void *data)
 	return INT2FIX(1);
       case Qtrue:
 	return INT2FIX(2);
+      default:
+	return Qnil;
     }
-    return Qnil;		/* not reached */
 }
 
+/*! Defines built-in variables */
 void
 ruby_prog_init(void)
 {
@@ -1807,6 +1854,13 @@ ruby_prog_init(void)
     rb_define_hooked_variable("$0", &rb_progname, 0, set_arg0);
     rb_define_hooked_variable("$PROGRAM_NAME", &rb_progname, 0, set_arg0);
 
+    /*
+     * ARGV contains the command line arguments used to run ruby with the
+     * first value containing the name of the executable.
+     *
+     * A library like OptionParser can be used to process command-line
+     * arguments.
+     */
     rb_define_global_const("ARGV", rb_argv);
 }
 
@@ -1824,7 +1878,7 @@ ruby_set_argv(int argc, char **argv)
 #endif
     rb_ary_clear(av);
     for (i = 0; i < argc; i++) {
-	VALUE arg = rb_external_str_new(argv[i], strlen(argv[i]));
+	VALUE arg = rb_external_str_new_cstr(argv[i]);
 
 	OBJ_FREEZE(arg);
 	rb_ary_push(av, arg);
@@ -1841,9 +1895,53 @@ ruby_process_options(int argc, char **argv)
     rb_argv0 = rb_str_new4(rb_progname);
     rb_gc_register_mark_object(rb_argv0);
     iseq = process_options(argc, argv, cmdline_options_init(&opt));
+
+#ifndef HAVE_SETPROCTITLE
+    {
+	extern void ruby_init_setproctitle(int argc, char *argv[]);
+	ruby_init_setproctitle(argc, argv);
+    }
+#endif
+
     return (void*)(struct RData*)iseq;
 }
 
+static void
+fill_standard_fds(void)
+{
+    int f0, f1, f2, fds[2];
+    struct stat buf;
+    f0 = fstat(0, &buf) == -1 && errno == EBADF;
+    f1 = fstat(1, &buf) == -1 && errno == EBADF;
+    f2 = fstat(2, &buf) == -1 && errno == EBADF;
+    if (f0) {
+        if (pipe(fds) == 0) {
+            close(fds[1]);
+            if (fds[0] != 0) {
+                dup2(fds[0], 0);
+                close(fds[0]);
+            }
+        }
+    }
+    if (f1 || f2) {
+        if (pipe(fds) == 0) {
+            close(fds[0]);
+            if (f1 && fds[1] != 1)
+                dup2(fds[1], 1);
+            if (f2 && fds[1] != 2)
+                dup2(fds[1], 2);
+            if (fds[1] != 1 && fds[1] != 2)
+                close(fds[1]);
+        }
+    }
+}
+
+/*! Initializes the process for ruby(1).
+ *
+ * This function assumes this process is ruby(1) and it has just started.
+ * Usually programs that embeds CRuby interpreter should not call this function,
+ * and should do their own initialization.
+ */
 void
 ruby_sysinit(int *argc, char ***argv)
 {
@@ -1853,10 +1951,8 @@ ruby_sysinit(int *argc, char ***argv)
 #endif
     origarg.argc = *argc;
     origarg.argv = *argv;
-#if !defined(PSTAT_SETCMD) && !defined(HAVE_SETPROCTITLE)
-    origarg.len = get_arglen(origarg.argc, origarg.argv);
-#endif
 #if defined(USE_DLN_A_OUT)
     dln_argv0 = origarg.argv[0];
 #endif
+    fill_standard_fds();
 }
