@@ -22,6 +22,7 @@
 #include "vm_opts.h"
 #include "id.h"
 #include "method.h"
+#include "atomic.h"
 
 #if   defined(_WIN32)
 #include "thread_win32.h"
@@ -31,6 +32,21 @@
 #error "unsupported thread type"
 #endif
 
+#ifndef ENABLE_VM_OBJSPACE
+#ifdef _WIN32
+/*
+ * TODO: object space independent st_table.
+ * socklist needs st_table in rb_w32_sysinit(), before object space
+ * initialization.
+ * It is too early now to change st_hash_type, since it breaks binary
+ * compatibility.
+ */
+#define ENABLE_VM_OBJSPACE 0
+#else
+#define ENABLE_VM_OBJSPACE 1
+#endif
+#endif
+
 #include <setjmp.h>
 #include <signal.h>
 
@@ -42,10 +58,14 @@
 
 #ifdef HAVE_STDARG_PROTOTYPES
 #include <stdarg.h>
-#define va_init_list(a,b) va_start(a,b)
+#define va_init_list(a,b) va_start((a),(b))
 #else
 #include <varargs.h>
-#define va_init_list(a,b) va_start(a)
+#define va_init_list(a,b) va_start((a))
+#endif
+
+#if defined(SIGSEGV) && defined(HAVE_SIGALTSTACK) && defined(SA_SIGINFO) && !defined(__NetBSD__)
+#define USE_SIGALTSTACK
 #endif
 
 /*****************/
@@ -72,6 +92,10 @@
 #endif
 #endif
 
+#ifdef __native_client__
+#undef OPT_DIRECT_THREADED_CODE
+#endif
+
 /* call threaded code */
 #if    OPT_CALL_THREADED_CODE
 #if    OPT_DIRECT_THREADED_CODE
@@ -91,42 +115,81 @@
 #define UNLIKELY(x) (x)
 #endif /* __GNUC__ >= 3 */
 
+#if __GNUC__ >= 3
+#define UNINITIALIZED_VAR(x) x = x
+#else
+#define UNINITIALIZED_VAR(x) x
+#endif
+
 typedef unsigned long rb_num_t;
 
 /* iseq data type */
 
 struct iseq_compile_data_ensure_node_stack;
 
-typedef struct rb_compile_option_struct {
-    int inline_const_cache;
-    int peephole_optimization;
-    int tailcall_optimization;
-    int specialized_instruction;
-    int operands_unification;
-    int instructions_unification;
-    int stack_caching;
-    int trace_instruction;
-    int debug_level;
-} rb_compile_option_t;
+typedef struct rb_compile_option_struct rb_compile_option_t;
 
 struct iseq_inline_cache_entry {
-    long  ic_vmstat;
+    VALUE ic_vmstat;
     VALUE ic_class;
-    VALUE ic_value;
-    rb_method_entry_t *ic_method;
-#define ic_index ic_vmstat
+    union {
+	VALUE value;
+	long index;
+    } ic_value;
 };
 
+/* to avoid warning */
+struct rb_thread_struct;
+struct rb_control_frame_struct;
+
+/* rb_call_info_t contains calling information including inline cache */
+typedef struct rb_call_info_struct {
+    /* fixed at compile time */
+    ID mid;
+    VALUE flag;
+    int orig_argc;
+    rb_iseq_t *blockiseq;
+
+    /* inline cache: keys */
+    VALUE vmstat;
+    VALUE klass;
+
+    /* inline cache: values */
+    const rb_method_entry_t *me;
+    VALUE defined_class;
+
+    /* temporary values for method calling */
+    int argc;
+    struct rb_block_struct *blockptr;
+    VALUE recv;
+    union {
+	int opt_pc; /* used by iseq */
+	long index; /* used by ivar */
+	int missing_reason; /* used by method_missing */
+	int inc_sp; /* used by cfunc */
+    } aux;
+
+    VALUE (*call)(struct rb_thread_struct *th, struct rb_control_frame_struct *cfp, struct rb_call_info_struct *ci);
+} rb_call_info_t;
+
 #if 1
 #define GetCoreDataFromValue(obj, type, ptr) do { \
-    ptr = (type*)DATA_PTR(obj); \
+    (ptr) = (type*)DATA_PTR(obj); \
 } while (0)
 #else
-#define GetCoreDataFromValue(obj, type, ptr) Data_Get_Struct(obj, type, ptr)
+#define GetCoreDataFromValue(obj, type, ptr) Data_Get_Struct((obj), type, (ptr))
 #endif
 
 #define GetISeqPtr(obj, ptr) \
-  GetCoreDataFromValue(obj, rb_iseq_t, ptr)
+  GetCoreDataFromValue((obj), rb_iseq_t, (ptr))
+
+typedef struct rb_iseq_location_struct {
+    VALUE path;
+    VALUE absolute_path;
+    VALUE base_label;
+    VALUE label;
+    size_t first_lineno;
+} rb_iseq_location_t;
 
 struct rb_iseq_struct;
 
@@ -135,18 +198,29 @@ struct rb_iseq_struct {
     /* static data */
     /***************/
 
-    VALUE type;          /* instruction sequence type */
-    VALUE name;	         /* String: iseq name */
-    VALUE filename;      /* file information where this sequence from */
-    VALUE *iseq;         /* iseq (insn number and openrads) */
+    enum iseq_type {
+	ISEQ_TYPE_TOP,
+	ISEQ_TYPE_METHOD,
+	ISEQ_TYPE_BLOCK,
+	ISEQ_TYPE_CLASS,
+	ISEQ_TYPE_RESCUE,
+	ISEQ_TYPE_ENSURE,
+	ISEQ_TYPE_EVAL,
+	ISEQ_TYPE_MAIN,
+	ISEQ_TYPE_DEFINED_GUARD
+    } type;              /* instruction sequence type */
+
+    rb_iseq_location_t location;
+
+    VALUE *iseq;         /* iseq (insn number and operands) */
     VALUE *iseq_encoded; /* encoded iseq */
     unsigned long iseq_size;
     VALUE mark_ary;	/* Array: includes operands which should be GC marked */
     VALUE coverage;     /* coverage array */
 
     /* insn info, must be freed */
-    struct iseq_insn_info_entry *insn_info_table;
-    size_t insn_info_size;
+    struct iseq_line_info_entry *line_info_table;
+    size_t line_info_size;
 
     ID *local_table;		/* must free */
     int local_table_size;
@@ -157,11 +231,14 @@ struct rb_iseq_struct {
     struct iseq_inline_cache_entry *ic_entries;
     int ic_size;
 
+    rb_call_info_t *callinfo_entries;
+    int callinfo_size;
+
     /**
      * argument information
      *
      *  def m(a1, a2, ..., aM,                    # mandatory
-     *        b1=(...), b2=(...), ..., bN=(...),  # optinal
+     *        b1=(...), b2=(...), ..., bN=(...),  # optional
      *        *c,                                 # rest
      *        d1, d2, ..., dO,                    # post
      *        &e)                                 # block
@@ -169,14 +246,14 @@ struct rb_iseq_struct {
      *
      *  argc           = M
      *  arg_rest       = M+N+1 // or -1 if no rest arg
-     *  arg_opts       = N
-     *  arg_opts_tbl   = [ (N entries) ]
+     *  arg_opts       = N+1   // or 0  if no optional arg
+     *  arg_opt_table  = [ (arg_opts entries) ]
      *  arg_post_len   = O // 0 if no post arguments
      *  arg_post_start = M+N+2
      *  arg_block      = M+N + 1 + O + 1 // -1 if no block arg
      *  arg_simple     = 0 if not simple arguments.
      *                 = 1 if no opt, rest, post, block.
-     *                 = 2 if ambiguos block parameter ({|a|}).
+     *                 = 2 if ambiguous block parameter ({|a|}).
      *  arg_size       = argument size.
      */
 
@@ -189,6 +266,10 @@ struct rb_iseq_struct {
     int arg_post_start;
     int arg_size;
     VALUE *arg_opt_table;
+    int arg_keyword;
+    int arg_keyword_check; /* if this is true, raise an ArgumentError when unknown keyword argument is passed */
+    int arg_keywords;
+    ID *arg_keyword_table;
 
     size_t stack_max; /* for stack overflow check */
 
@@ -230,16 +311,28 @@ enum ruby_special_exceptions {
     ruby_error_reenter,
     ruby_error_nomemory,
     ruby_error_sysstack,
+    ruby_error_closed_stream,
     ruby_special_error_count
 };
 
 #define GetVMPtr(obj, ptr) \
-  GetCoreDataFromValue(obj, rb_vm_t, ptr)
+  GetCoreDataFromValue((obj), rb_vm_t, (ptr))
+
+#if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
+struct rb_objspace;
+void rb_objspace_free(struct rb_objspace *);
+#endif
+
+typedef struct rb_hook_list_struct {
+    struct rb_event_hook_struct *hooks;
+    rb_event_flag_t events;
+    int need_clean;
+} rb_hook_list_t;
 
 typedef struct rb_vm_struct {
     VALUE self;
 
-    rb_thread_lock_t global_vm_lock;
+    rb_global_vm_lock_t gvl;
 
     struct rb_thread_struct *main_thread;
     struct rb_thread_struct *running_thread;
@@ -248,8 +341,9 @@ typedef struct rb_vm_struct {
     VALUE thgroup_default;
 
     int running;
+    int inhibit_thread_creation;
     int thread_abort_on_exception;
-    unsigned long trace_flag;
+    int trace_running;
     volatile int sleeper;
 
     /* object management */
@@ -270,42 +364,61 @@ typedef struct rb_vm_struct {
     } trap_list[RUBY_NSIG];
 
     /* hook */
-    rb_event_hook_t *event_hooks;
+    rb_hook_list_t event_hooks;
 
     int src_encoding_index;
 
     VALUE verbose, debug, progname;
     VALUE coverages;
 
+    struct unlinked_method_entry_list_entry *unlinked_method_entry_list;
+
 #if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
     struct rb_objspace *objspace;
 #endif
+
+    /*
+     * @shyouhei notes that this is not for storing normal Ruby
+     * objects so do *NOT* mark this when you GC.
+     */
+    struct RArray at_exit;
+
+    VALUE *defined_strings;
 } rb_vm_t;
 
-typedef struct {
+#ifndef VM_DEBUG_BP_CHECK
+#define VM_DEBUG_BP_CHECK 1
+#endif
+
+typedef struct rb_control_frame_struct {
     VALUE *pc;			/* cfp[0] */
     VALUE *sp;			/* cfp[1] */
-    VALUE *bp;			/* cfp[2] */
-    rb_iseq_t *iseq;		/* cfp[3] */
-    VALUE flag;			/* cfp[4] */
-    VALUE self;			/* cfp[5] / block[0] */
-    VALUE *lfp;			/* cfp[6] / block[1] */
-    VALUE *dfp;			/* cfp[7] / block[2] */
-    rb_iseq_t *block_iseq;	/* cfp[8] / block[3] */
-    VALUE proc;			/* cfp[9] / block[4] */
-    const rb_method_entry_t *me;/* cfp[10] */
+    rb_iseq_t *iseq;		/* cfp[2] */
+    VALUE flag;			/* cfp[3] */
+    VALUE self;			/* cfp[4] / block[0] */
+    VALUE klass;		/* cfp[5] / block[1] */
+    VALUE *ep;			/* cfp[6] / block[2] */
+    rb_iseq_t *block_iseq;	/* cfp[7] / block[3] */
+    VALUE proc;			/* cfp[8] / block[4] */
+    const rb_method_entry_t *me;/* cfp[9] */
+
+#if VM_DEBUG_BP_CHECK
+    VALUE *bp_check;		/* cfp[10] */
+#endif
 } rb_control_frame_t;
 
 typedef struct rb_block_struct {
     VALUE self;			/* share with method frame if it's only block */
-    VALUE *lfp;			/* share with method frame if it's only block */
-    VALUE *dfp;			/* share with method frame if it's only block */
+    VALUE klass;		/* share with method frame if it's only block */
+    VALUE *ep;			/* share with method frame if it's only block */
     rb_iseq_t *iseq;
     VALUE proc;
 } rb_block_t;
 
+extern const rb_data_type_t ruby_threadptr_data_type;
+
 #define GetThreadPtr(obj, ptr) \
-  GetCoreDataFromValue(obj, rb_thread_t, ptr)
+    TypedData_Get_Struct((obj), rb_thread_t, &ruby_threadptr_data_type, (ptr))
 
 enum rb_thread_status {
     THREAD_TO_KILL,
@@ -324,13 +437,10 @@ struct rb_vm_tag {
     struct rb_vm_tag *prev;
 };
 
-struct rb_vm_trap_tag {
-    struct rb_vm_trap_tag *prev;
+struct rb_vm_protect_tag {
+    struct rb_vm_protect_tag *prev;
 };
 
-#define RUBY_VM_VALUE_CACHE_SIZE 0x1000
-#define USE_VALUE_CACHE 0
-
 struct rb_unblock_callback {
     rb_unblock_function_t *func;
     void *arg;
@@ -338,8 +448,13 @@ struct rb_unblock_callback {
 
 struct rb_mutex_struct;
 
-typedef struct rb_thread_struct
-{
+#ifdef MINSIGSTKSZ
+#define ALT_STACK_SIZE (MINSIGSTKSZ*2)
+#else
+#define ALT_STACK_SIZE (4*1024)
+#endif
+
+typedef struct rb_thread_struct {
     VALUE self;
     rb_vm_t *vm;
 
@@ -353,10 +468,19 @@ typedef struct rb_thread_struct
 
     /* passing state */
     int state;
+    int yielding;
+
+    int waiting_fd;
 
     /* for rb_iterate */
     const rb_block_t *passed_block;
 
+    /* for bmethod */
+    const rb_method_entry_t *passed_me;
+
+    /* for cfunc */
+    rb_call_info_t *passed_ci;
+
     /* for load(true) */
     VALUE top_self;
     VALUE top_wrapper;
@@ -364,14 +488,13 @@ typedef struct rb_thread_struct
     /* eval env */
     rb_block_t *base_block;
 
-    VALUE *local_lfp;
-    VALUE local_svar;
+    VALUE *root_lep;
+    VALUE root_svar;
 
     /* thread control */
     rb_thread_id_t thread_id;
     enum rb_thread_status status;
     int priority;
-    int slice;
 
     native_thread_data_t native_thread_data;
     void *blocking_region_buffer;
@@ -379,29 +502,44 @@ typedef struct rb_thread_struct
     VALUE thgroup;
     VALUE value;
 
+    /* temporary place of errinfo */
     VALUE errinfo;
-    VALUE thrown_errinfo;
-    int exec_signal;
 
-    int interrupt_flag;
+    /* temporary place of retval on OPT_CALL_THREADED_CODE */
+#if OPT_CALL_THREADED_CODE
+    VALUE retval;
+#endif
+
+    /* async errinfo queue */
+    VALUE async_errinfo_queue;
+    int async_errinfo_queue_checked;
+    VALUE async_errinfo_mask_stack;
+
+    rb_atomic_t interrupt_flag;
     rb_thread_lock_t interrupt_lock;
     struct rb_unblock_callback unblock;
     VALUE locking_mutex;
     struct rb_mutex_struct *keeping_mutexes;
-    int transition_for_lock;
 
     struct rb_vm_tag *tag;
-    struct rb_vm_trap_tag *trap_tag;
+    struct rb_vm_protect_tag *protect_tag;
 
+    /*! Thread-local state of evaluation context.
+     *
+     *  If negative, this thread is evaluating the main program.
+     *  If positive, this thread is evaluating a program under Kernel::eval
+     *  family.
+     */
     int parse_in_eval;
+
+    /*! Thread-local state of compiling context.
+     *
+     * If non-zero, the parser does not automatically print error messages to
+     * stderr. */
     int mild_compile_error;
 
     /* storage */
     st_table *local_storage;
-#if USE_VALUE_CACHE
-    VALUE value_cache[RUBY_VM_VALUE_CACHE_SIZE + 1];
-    VALUE *value_cache_ptr;
-#endif
 
     struct rb_thread_struct *join_list_next;
     struct rb_thread_struct *join_list_head;
@@ -426,9 +564,8 @@ typedef struct rb_thread_struct
     VALUE stat_insn_usage;
 
     /* tracer */
-    rb_event_hook_t *event_hooks;
-    rb_event_flag_t event_flags;
-    int tracing;
+    rb_hook_list_t event_hooks;
+    int trace_running;
 
     /* fiber */
     VALUE fiber;
@@ -438,30 +575,47 @@ typedef struct rb_thread_struct
     /* misc */
     int method_missing_reason;
     int abort_on_exception;
+#ifdef USE_SIGALTSTACK
+    void *altstack;
+#endif
+    unsigned long running_time_us;
 } rb_thread_t;
 
 /* iseq.c */
-VALUE rb_iseq_new(NODE*, VALUE, VALUE, VALUE, VALUE);
-VALUE rb_iseq_new_top(NODE *node, VALUE name, VALUE filename, VALUE parent);
-VALUE rb_iseq_new_main(NODE *node, VALUE filename);
-VALUE rb_iseq_new_with_bopt(NODE*, VALUE, VALUE, VALUE, VALUE, VALUE);
-VALUE rb_iseq_new_with_opt(NODE*, VALUE, VALUE, VALUE, VALUE, const rb_compile_option_t*);
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility push(default)
+#endif
+
+/* node -> iseq */
+VALUE rb_iseq_new(NODE*, VALUE, VALUE, VALUE, VALUE, enum iseq_type);
+VALUE rb_iseq_new_top(NODE *node, VALUE name, VALUE path, VALUE absolute_path, VALUE parent);
+VALUE rb_iseq_new_main(NODE *node, VALUE path, VALUE absolute_path);
+VALUE rb_iseq_new_with_bopt(NODE*, VALUE, VALUE, VALUE, VALUE, VALUE, enum iseq_type, VALUE);
+VALUE rb_iseq_new_with_opt(NODE*, VALUE, VALUE, VALUE, VALUE, VALUE, enum iseq_type, const rb_compile_option_t*);
+
+/* src -> iseq */
 VALUE rb_iseq_compile(VALUE src, VALUE file, VALUE line);
+VALUE rb_iseq_compile_on_base(VALUE src, VALUE file, VALUE line, rb_block_t *base_block);
+VALUE rb_iseq_compile_with_option(VALUE src, VALUE file, VALUE absolute_path, VALUE line, rb_block_t *base_block, VALUE opt);
+
 VALUE rb_iseq_disasm(VALUE self);
-VALUE rb_iseq_disasm_insn(VALUE str, VALUE *iseqval, int pos, rb_iseq_t *iseq, VALUE child);
+int rb_iseq_disasm_insn(VALUE str, VALUE *iseqval, size_t pos, rb_iseq_t *iseq, VALUE child);
 const char *ruby_node_name(int node);
-int rb_iseq_first_lineno(rb_iseq_t *iseq);
+int rb_iseq_first_lineno(const rb_iseq_t *iseq);
 
 RUBY_EXTERN VALUE rb_cISeq;
 RUBY_EXTERN VALUE rb_cRubyVM;
 RUBY_EXTERN VALUE rb_cEnv;
 RUBY_EXTERN VALUE rb_mRubyVMFrozenCore;
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility pop
+#endif
 
 /* each thread has this size stack : 128KB */
 #define RUBY_VM_THREAD_STACK_SIZE (128 * 1024)
 
 #define GetProcPtr(obj, ptr) \
-  GetCoreDataFromValue(obj, rb_proc_t, ptr)
+  GetCoreDataFromValue((obj), rb_proc_t, (ptr))
 
 typedef struct {
     rb_block_t block;
@@ -474,7 +628,7 @@ typedef struct {
 } rb_proc_t;
 
 #define GetEnvPtr(obj, ptr) \
-  GetCoreDataFromValue(obj, rb_env_t, ptr)
+  GetCoreDataFromValue((obj), rb_env_t, (ptr))
 
 typedef struct {
     VALUE *env;
@@ -485,29 +639,44 @@ typedef struct {
 } rb_env_t;
 
 #define GetBindingPtr(obj, ptr) \
-  GetCoreDataFromValue(obj, rb_binding_t, ptr)
+  GetCoreDataFromValue((obj), rb_binding_t, (ptr))
 
 typedef struct {
     VALUE env;
+    VALUE path;
+    unsigned short first_lineno;
 } rb_binding_t;
 
 /* used by compile time and send insn */
-#define VM_CALL_ARGS_SPLAT_BIT     (0x01 << 1)
-#define VM_CALL_ARGS_BLOCKARG_BIT  (0x01 << 2)
-#define VM_CALL_FCALL_BIT          (0x01 << 3)
-#define VM_CALL_VCALL_BIT          (0x01 << 4)
-#define VM_CALL_TAILCALL_BIT       (0x01 << 5)
-#define VM_CALL_TAILRECURSION_BIT  (0x01 << 6)
-#define VM_CALL_SUPER_BIT          (0x01 << 7)
 
-#define VM_SPECIAL_OBJECT_VMCORE   0x01
-#define VM_SPECIAL_OBJECT_CBASE    0x02
+enum vm_check_match_type {
+    VM_CHECKMATCH_TYPE_WHEN = 1,
+    VM_CHECKMATCH_TYPE_CASE = 2,
+    VM_CHECKMATCH_TYPE_RESCUE = 3
+};
+
+#define VM_CHECKMATCH_TYPE_MASK   0x03
+#define VM_CHECKMATCH_ARRAY       0x04
+
+#define VM_CALL_ARGS_SPLAT      (0x01 << 1) /* m(*args) */
+#define VM_CALL_ARGS_BLOCKARG   (0x01 << 2) /* m(&block) */
+#define VM_CALL_FCALL           (0x01 << 3) /* m(...) */
+#define VM_CALL_VCALL           (0x01 << 4) /* m */
+#define VM_CALL_TAILCALL        (0x01 << 5) /* located at tail position */
+#define VM_CALL_SUPER           (0x01 << 6) /* super */
+#define VM_CALL_OPT_SEND        (0x01 << 7) /* internal flag */
+#define VM_CALL_ARGS_SKIP_SETUP (0x01 << 8) /* (flag & (SPLAT|BLOCKARG)) && blockiseq == 0 */
+
+enum vm_special_object_type {
+    VM_SPECIAL_OBJECT_VMCORE = 1,
+    VM_SPECIAL_OBJECT_CBASE,
+    VM_SPECIAL_OBJECT_CONST_BASE
+};
 
 #define VM_FRAME_MAGIC_METHOD 0x11
 #define VM_FRAME_MAGIC_BLOCK  0x21
 #define VM_FRAME_MAGIC_CLASS  0x31
 #define VM_FRAME_MAGIC_TOP    0x41
-#define VM_FRAME_MAGIC_FINISH 0x51
 #define VM_FRAME_MAGIC_CFUNC  0x61
 #define VM_FRAME_MAGIC_PROC   0x71
 #define VM_FRAME_MAGIC_IFUNC  0x81
@@ -520,12 +689,15 @@ typedef struct {
 
 /* other frame flag */
 #define VM_FRAME_FLAG_PASSED 0x0100
+#define VM_FRAME_FLAG_FINISH 0x0200
+#define VM_FRAME_TYPE_FINISH_P(cfp) (((cfp)->flag & VM_FRAME_FLAG_FINISH) != 0)
 
 #define RUBYVM_CFUNC_FRAME_P(cfp) \
   (VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_CFUNC)
 
 /* inline cache */
 typedef struct iseq_inline_cache_entry *IC;
+typedef rb_call_info_t *CALL_INFO;
 
 void rb_vm_change_state(void);
 
@@ -539,11 +711,33 @@ typedef rb_control_frame_t *
   (FUNC_FASTCALL(*rb_insn_func_t))(rb_thread_t *, rb_control_frame_t *);
 
 #define GC_GUARDED_PTR(p)     ((VALUE)((VALUE)(p) | 0x01))
-#define GC_GUARDED_PTR_REF(p) ((void *)(((VALUE)p) & ~0x03))
-#define GC_GUARDED_PTR_P(p)   (((VALUE)p) & 0x01)
+#define GC_GUARDED_PTR_REF(p) ((void *)(((VALUE)(p)) & ~0x03))
+#define GC_GUARDED_PTR_P(p)   (((VALUE)(p)) & 0x01)
 
-#define RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp) (cfp+1)
-#define RUBY_VM_NEXT_CONTROL_FRAME(cfp) (cfp-1)
+/*
+ * block frame:
+ *  ep[ 0]: prev frame
+ *  ep[-1]: CREF (for *_eval)
+ *
+ * method frame:
+ *  ep[ 0]: block pointer (ptr | VM_ENVVAL_BLOCK_PTR_FLAG)
+ */
+
+#define VM_ENVVAL_BLOCK_PTR_FLAG 0x02
+#define VM_ENVVAL_BLOCK_PTR(v)     (GC_GUARDED_PTR(v) | VM_ENVVAL_BLOCK_PTR_FLAG)
+#define VM_ENVVAL_BLOCK_PTR_P(v)   ((v) & VM_ENVVAL_BLOCK_PTR_FLAG)
+#define VM_ENVVAL_PREV_EP_PTR(v)   ((VALUE)GC_GUARDED_PTR(v))
+#define VM_ENVVAL_PREV_EP_PTR_P(v) (!(VM_ENVVAL_BLOCK_PTR_P(v)))
+
+#define VM_EP_PREV_EP(ep)   ((VALUE *)GC_GUARDED_PTR_REF((ep)[0]))
+#define VM_EP_BLOCK_PTR(ep) ((rb_block_t *)GC_GUARDED_PTR_REF((ep)[0]))
+#define VM_EP_LEP_P(ep)     VM_ENVVAL_BLOCK_PTR_P((ep)[0])
+
+VALUE *rb_vm_ep_local_ep(VALUE *ep);
+rb_block_t *rb_vm_control_frame_block_ptr(rb_control_frame_t *cfp);
+
+#define RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp) ((cfp)+1)
+#define RUBY_VM_NEXT_CONTROL_FRAME(cfp) ((cfp)-1)
 #define RUBY_VM_END_CONTROL_FRAME(th) \
   ((rb_control_frame_t *)((th)->stack + (th)->stack_size))
 #define RUBY_VM_VALID_CONTROL_FRAME_P(cfp, ecfp) \
@@ -553,13 +747,12 @@ typedef rb_control_frame_t *
 
 #define RUBY_VM_IFUNC_P(ptr)        (BUILTIN_TYPE(ptr) == T_NODE)
 #define RUBY_VM_NORMAL_ISEQ_P(ptr) \
-  (ptr && !RUBY_VM_IFUNC_P(ptr))
-
-#define RUBY_VM_CLASS_SPECIAL_P(ptr) (((VALUE)(ptr)) & 0x02)
+  ((ptr) && !RUBY_VM_IFUNC_P(ptr))
 
 #define RUBY_VM_GET_BLOCK_PTR_IN_CFP(cfp) ((rb_block_t *)(&(cfp)->self))
 #define RUBY_VM_GET_CFP_FROM_BLOCK_PTR(b) \
-  ((rb_control_frame_t *)((VALUE *)(b) - 5))
+  ((rb_control_frame_t *)((VALUE *)(b) - 4))
+/* magic number `4' is depend on rb_control_frame_t layout. */
 
 /* VM related object allocate functions */
 VALUE rb_thread_alloc(VALUE klass);
@@ -567,83 +760,152 @@ VALUE rb_proc_alloc(VALUE klass);
 
 /* for debug */
 extern void rb_vmdebug_stack_dump_raw(rb_thread_t *, rb_control_frame_t *);
+extern void rb_vmdebug_debug_print_pre(rb_thread_t *th, rb_control_frame_t *cfp);
+extern void rb_vmdebug_debug_print_post(rb_thread_t *th, rb_control_frame_t *cfp);
+
 #define SDR() rb_vmdebug_stack_dump_raw(GET_THREAD(), GET_THREAD()->cfp)
 #define SDR2(cfp) rb_vmdebug_stack_dump_raw(GET_THREAD(), (cfp))
 void rb_vm_bugreport(void);
 
 /* functions about thread/vm execution */
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility push(default)
+#endif
 VALUE rb_iseq_eval(VALUE iseqval);
 VALUE rb_iseq_eval_main(VALUE iseqval);
-void rb_enable_interrupt(void);
-void rb_disable_interrupt(void);
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility pop
+#endif
 int rb_thread_method_id_and_class(rb_thread_t *th, ID *idp, VALUE *klassp);
 
-VALUE rb_vm_invoke_proc(rb_thread_t *th, rb_proc_t *proc, VALUE self,
+VALUE rb_vm_invoke_proc(rb_thread_t *th, rb_proc_t *proc,
 			int argc, const VALUE *argv, const rb_block_t *blockptr);
 VALUE rb_vm_make_proc(rb_thread_t *th, const rb_block_t *block, VALUE klass);
 VALUE rb_vm_make_env_object(rb_thread_t *th, rb_control_frame_t *cfp);
+void rb_vm_rewrite_ep_in_errinfo(rb_thread_t *th, rb_control_frame_t *cfp);
+void rb_vm_inc_const_missing_count(void);
+void rb_vm_gvl_destroy(rb_vm_t *vm);
+VALUE rb_vm_call(rb_thread_t *th, VALUE recv, VALUE id, int argc,
+		 const VALUE *argv, const rb_method_entry_t *me,
+		 VALUE defined_class);
+void rb_unlink_method_entry(rb_method_entry_t *me);
+void rb_gc_mark_unlinked_live_method_entries(void *pvm);
+
+void rb_thread_start_timer_thread(void);
+void rb_thread_stop_timer_thread(int);
+void rb_thread_reset_timer_thread(void);
+void rb_thread_wakeup_timer_thread(void);
 
-void *rb_thread_call_with_gvl(void *(*func)(void *), void *data1);
 int ruby_thread_has_gvl_p(void);
-VALUE rb_make_backtrace(void);
-typedef int rb_backtrace_iter_func(void *, const char *, int, const char *);
-VALUE rb_backtrace_each(rb_backtrace_iter_func *iter, void *arg);
+typedef int rb_backtrace_iter_func(void *, VALUE, int, VALUE);
 rb_control_frame_t *rb_vm_get_ruby_level_next_cfp(rb_thread_t *th, rb_control_frame_t *cfp);
+int rb_vm_get_sourceline(const rb_control_frame_t *);
+VALUE rb_name_err_mesg_new(VALUE obj, VALUE mesg, VALUE recv, VALUE method);
+void rb_vm_stack_to_heap(rb_thread_t *th);
+void ruby_thread_init_stack(rb_thread_t *th);
 
 NOINLINE(void rb_gc_save_machine_context(rb_thread_t *));
+void rb_gc_mark_machine_stack(rb_thread_t *th);
+
+int rb_autoloading_value(VALUE mod, ID id, VALUE* value);
 
 #define sysstack_error GET_VM()->special_exceptions[ruby_error_sysstack]
 
-VALUE rb_str_resurrect(VALUE str);
-VALUE rb_ary_resurrect(VALUE ary);
-
 /* for thread */
 
 #if RUBY_VM_THREAD_MODEL == 2
-RUBY_EXTERN rb_thread_t *ruby_current_thread;
+extern rb_thread_t *ruby_current_thread;
 extern rb_vm_t *ruby_current_vm;
+extern rb_event_flag_t ruby_vm_event_flags;
 
 #define GET_VM() ruby_current_vm
-#define GET_THREAD() ruby_current_thread
+
+#ifndef OPT_CALL_CFUNC_WITHOUT_FRAME
+#define OPT_CALL_CFUNC_WITHOUT_FRAME 0
+#endif
+
+static inline rb_thread_t *
+GET_THREAD(void)
+{
+    rb_thread_t *th = ruby_current_thread;
+#if OPT_CALL_CFUNC_WITHOUT_FRAME
+    if (UNLIKELY(th->passed_ci != 0)) {
+	void vm_call_cfunc_push_frame(rb_thread_t *th);
+	vm_call_cfunc_push_frame(th);
+    }
+#endif
+    return th;
+}
+
 #define rb_thread_set_current_raw(th) (void)(ruby_current_thread = (th))
 #define rb_thread_set_current(th) do { \
+    if ((th)->vm->running_thread != (th)) { \
+	(th)->vm->running_thread->running_time_us = 0; \
+    } \
     rb_thread_set_current_raw(th); \
-    th->vm->running_thread = th; \
+    (th)->vm->running_thread = (th); \
 } while (0)
 
 #else
 #error "unsupported thread model"
 #endif
 
-#define RUBY_VM_SET_INTERRUPT(th) ((th)->interrupt_flag |= 0x02)
-#define RUBY_VM_SET_TIMER_INTERRUPT(th) ((th)->interrupt_flag |= 0x01)
-#define RUBY_VM_SET_FINALIZER_INTERRUPT(th) ((th)->interrupt_flag |= 0x04)
+#define RUBY_VM_SET_TIMER_INTERRUPT(th)		ATOMIC_OR((th)->interrupt_flag, 0x01)
+#define RUBY_VM_SET_INTERRUPT(th)		ATOMIC_OR((th)->interrupt_flag, 0x02)
+#define RUBY_VM_SET_FINALIZER_INTERRUPT(th)	ATOMIC_OR((th)->interrupt_flag, 0x04)
 #define RUBY_VM_INTERRUPTED(th) ((th)->interrupt_flag & 0x02)
 
+int rb_signal_buff_size(void);
+void rb_signal_exec(rb_thread_t *th, int sig);
+void rb_threadptr_check_signal(rb_thread_t *mth);
 void rb_threadptr_signal_raise(rb_thread_t *th, int sig);
 void rb_threadptr_signal_exit(rb_thread_t *th);
-void rb_threadptr_execute_interrupts(rb_thread_t *);
+void rb_threadptr_execute_interrupts(rb_thread_t *, int);
+void rb_threadptr_interrupt(rb_thread_t *th);
+void rb_threadptr_unlock_all_locking_mutexes(rb_thread_t *th);
+void rb_threadptr_async_errinfo_clear(rb_thread_t *th);
+void rb_threadptr_async_errinfo_enque(rb_thread_t *th, VALUE v);
+int rb_threadptr_async_errinfo_active_p(rb_thread_t *th);
 
-#define RUBY_VM_CHECK_INTS_TH(th) do { \
-  if (UNLIKELY(th->interrupt_flag)) { \
-    rb_threadptr_execute_interrupts(th); \
-  } \
+void rb_thread_lock_unlock(rb_thread_lock_t *);
+void rb_thread_lock_destroy(rb_thread_lock_t *);
+
+#define RUBY_VM_CHECK_INTS_BLOCKING(th) do { \
+    if (UNLIKELY((th)->interrupt_flag)) { \
+	rb_threadptr_execute_interrupts(th, 1); \
+    } \
 } while (0)
 
-#define RUBY_VM_CHECK_INTS() \
-  RUBY_VM_CHECK_INTS_TH(GET_THREAD())
+#define RUBY_VM_CHECK_INTS(th) do { \
+    if (UNLIKELY((th)->interrupt_flag)) { \
+	rb_threadptr_execute_interrupts(th, 0); \
+    } \
+} while (0)
 
 /* tracer */
 void
 rb_threadptr_exec_event_hooks(rb_thread_t *th, rb_event_flag_t flag, VALUE self, ID id, VALUE klass);
 
 #define EXEC_EVENT_HOOK(th, flag, self, id, klass) do { \
-    rb_event_flag_t wait_event__ = th->event_flags; \
-    if (UNLIKELY(wait_event__)) { \
-	if (wait_event__ & (flag | RUBY_EVENT_VM)) { \
-	    rb_threadptr_exec_event_hooks(th, flag, self, id, klass); \
+    if (UNLIKELY(ruby_vm_event_flags & (flag))) { \
+	if (((th)->event_hooks.events | (th)->vm->event_hooks.events) & (flag)) { \
+	    rb_threadptr_exec_event_hooks((th), (flag), (self), (id), (klass)); \
 	} \
     } \
 } while (0)
 
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility push(default)
+#endif
+
+int rb_thread_check_trap_pending(void);
+
+extern VALUE rb_get_coverages(void);
+extern void rb_set_coverages(VALUE);
+extern void rb_reset_coverages(void);
+
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility pop
+#endif
+
 #endif /* RUBY_VM_CORE_H */
